
/*HEADER---------------------------------------------------
* $Id: linked_list.h,v 1.1 2004/09/17 05:35:36 gjhurlbu Exp $
*
* Copyright 2004 Gavin Hurlbut
* All rights reserved
*
* Comments :
*
* Revision History:
* $Log: linked_list.h,v $
* Revision 1.1  2004/09/17 05:35:36  gjhurlbu
* - changed to using linked lists rather than arrays for storing encoded frames
*
*
*
*--------------------------------------------------------*/
#ifndef linked_list_h_
#define linked_list_h_

#include "environment.h"

/* CVS generated ID string (optional for h files) */
static char linked_list_h_ident[] _UNUSED_ = 
    "$Id: linked_list.h,v 1.1 2004/09/17 05:35:36 gjhurlbu Exp $";


struct _LinkedList_t;
struct _LinkedListItem_t;

/* This *MUST* be embedded at the BEGINNING of any type that will be used in a 
 * linked list 
 */
typedef struct _LinkedListItem_t
{
    struct _LinkedListItem_t *prev;
    struct _LinkedListItem_t *next;
    struct _LinkedList_t     *list;
} LinkedListItem_t;

typedef struct _LinkedList_t
{
    LinkedListItem_t *head;
    LinkedListItem_t *tail;
    pthread_mutex_t  mutex;
} LinkedList_t;

typedef enum
{
    UNLOCKED,
    LOCKED
} LinkedListLocked_t;

typedef enum
{
    AT_HEAD,
    AT_TAIL
} LinkedListLoc_t;


#ifdef __cplusplus
extern "C" {
#endif

/* Function Prototypes */
LinkedList_t *LinkedListCreate( void );
void LinkedListDestroy( LinkedList_t *list );
void LinkedListLock( LinkedList_t *list );
void LinkedListUnlock( LinkedList_t *list );
void LinkedListAdd( LinkedList_t *list, LinkedListItem_t *item, 
                    LinkedListLocked_t locked, LinkedListLoc_t location );
void LinkedListRemove( LinkedList_t *list, LinkedListItem_t *item, 
                       LinkedListLocked_t locked );

#ifdef __cplusplus
}
#endif

#endif
