
/*HEADER---------------------------------------------------
* $Id: reader.c,v 1.3 2004/09/15 05:59:46 gjhurlbu Exp $
*
* Copyright 2004 Gavin Hurlbut
* All rights reserved
*
* Comments :
*
* Revision History:
* $Log: reader.c,v $
* Revision 1.3  2004/09/15 05:59:46  gjhurlbu
* - added graceful aborting
* - added a GOP length queue in the reencoder
* - added status update in the writer
*
* Revision 1.2  2004/09/13 05:30:40  gjhurlbu
* - added ident string
*
* Revision 1.1  2004/09/13 05:29:19  gjhurlbu
* - converted to threaded
*
*
*
*----------------------------------------------------------
*/

#include "environment.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>

#include "format.h"
#include "nuvread.h"
#include "queue.h"

/* CVS generated ID string */
static char ident[] _UNUSED_ = 
    "$Id: reader.c,v 1.3 2004/09/15 05:59:46 gjhurlbu Exp $";


extern QueueObject_t *DecodeQueue;
extern bool GlobalAbort;

void *ReaderThread( void *arg )
{
    char           *filename;
    nuvFile        *nv;
    nuvFrame       *frame;

    filename = (char *)arg;

    nv = nuv_init( filename, 0 );
    if( !nv ) {
        fprintf( stderr, "Unable to open file" );
        return( NULL );
    }

    frame = (nuvFrame *)malloc(sizeof(nuvFrame));
    if( !frame ) {
        fprintf( stderr, "Can't pass config to the reencode thread" );
        return( NULL );
    }

    frame->header.frametype = 'I';
    frame->length = sizeof( nuvFile );
    frame->data = (char *)nv;
    QueueEnqueueItem( DecodeQueue, (QueueItem_t)frame );

    while( (frame = nuv_read_frame( nv, "AVS" )) ) {
        if( GlobalAbort ) {
            return( NULL );
        }
        QueueEnqueueItem( DecodeQueue, (QueueItem_t)frame );
    }

#if 0
    printf( "\nTotal NUV Frame Count Read: %lld\n\n", nv->frameCount );
#endif

    nuv_free_file(nv);

    frame = (nuvFrame *)malloc(sizeof(nuvFrame));
    if( !frame ) {
        fprintf( stderr, "Can't pass end frame to the reencode thread" );
        return( NULL );
    }

    frame->header.frametype = 'E';
    frame->length = 0;
    frame->data = NULL;
    QueueEnqueueItem( DecodeQueue, (QueueItem_t)frame );

    return( NULL );
}


/*
 * vim:ts=4:sw=4:ai:et:si:sts=4
 */
