
/*HEADER---------------------------------------------------
* $Id: reencode.c,v 1.10 2004/09/19 21:34:24 gjhurlbu Exp $
*
* Copyright 2004 Gavin Hurlbut
* All rights reserved
*
* Comments :
*
* Revision History:
* $Log: reencode.c,v $
* Revision 1.10  2004/09/19 21:34:24  gjhurlbu
* - doesn't bork the audio now, but doesn't work (still)
*
* Revision 1.9  2004/09/19 18:27:50  gjhurlbu
* - tries to reencode audio, but borks it
*
* Revision 1.8  2004/09/17 23:47:10  gjhurlbu
* - Changed THRESHOLD to be based on the VideoFrameTime
*
* Revision 1.7  2004/09/17 05:35:36  gjhurlbu
* - changed to using linked lists rather than arrays for storing encoded frames
*
* Revision 1.6  2004/09/17 03:32:59  gjhurlbu
* - updated display to update once a second
*
* Revision 1.5  2004/09/16 05:35:26  gjhurlbu
* - renumber audio and video frames if they are more than 100ms after the
*   previous frame
*
* Revision 1.4  2004/09/16 03:59:17  gjhurlbu
* - added offset calculations to resync cut files
*
* Revision 1.3  2004/09/15 05:59:46  gjhurlbu
* - added graceful aborting
* - added a GOP length queue in the reencoder
* - added status update in the writer
*
* Revision 1.2  2004/09/13 05:30:40  gjhurlbu
* - added ident string
*
* Revision 1.1  2004/09/13 05:29:19  gjhurlbu
* - converted to threaded
*
*
*
*----------------------------------------------------------
*/

#include "environment.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>

#include "format.h"
#include "nuvread.h"
#include "queue.h"
#include "linked_list.h"
#include <lame.h>
#include "avcodec.h"

/* CVS generated ID string */
static char ident[] _UNUSED_ = 
    "$Id: reencode.c,v 1.10 2004/09/19 21:34:24 gjhurlbu Exp $";

extern QueueObject_t *DecodeQueue;
extern QueueObject_t *WritingQueue;
extern bool GlobalAbort;

typedef struct {
    LinkedListItem_t    linkage;
    nuvFrame           *frame;
} linkedFrame;

static LinkedList_t *videoList;
static LinkedList_t *audioList;

static nuvFile        *nv;
static nuvFrame       *frame;
static nuvFrame       *audFrame;
static int             timecode;
static int             VideoOffset = 0;
static int             VideoPrev;
static int             VideoLastPrev;
static int             VideoFrameTime;
static linkedFrame    *item;
static LinkedListItem_t *currVid, *currAud;
static lame_global_flags *gf, *gfOut;
static int             lameret;
static short int      *pcmlbuffer;
static short int      *pcmrbuffer;
static short int      *pcmtemp;
static short int      *temp;
static int             pcmlen;
static int             bufOffset;
static float           AudTime;
static float           VidTime;
static float           Ratio;
static struct AVResampleContext *resample;
static int             consumed;
static int             count;
static int             mp3size;

void *ReencodeThread( void *arg )
{
    int             done;
    int             syncFrame = 0;
    int             synched = 0;
    int             audRaw = 0;


    done = FALSE;

    videoList = LinkedListCreate();
    audioList = LinkedListCreate();

    if( !videoList || !audioList ) {
        fprintf( stderr, "Couldn't create linked list\n" );
        GlobalAbort = TRUE;
        return( NULL );
    }

    gf = lame_init();
    gfOut = lame_init();

    while( !done ) {
        frame = (nuvFrame *)QueueDequeueItem( DecodeQueue, -1 );
        if( !frame || GlobalAbort ) {
            return( NULL );
        }

        switch( frame->header.frametype ) {
        case 'I':
            nv = (nuvFile *)frame->data;

            VideoFrameTime = (int)(1000.0 * (float)nv->fps.denom /
                                   (float)nv->fps.num);
            VideoPrev = -1 * VideoFrameTime;

            lame_set_bWriteVbrTag(gf, 0);
            lame_set_quality(gf, nv->audio_quality);
            lame_set_brate(gf, nv->audio_bitrate_closest);
            lame_set_mode(gf, nv->audio_channels == 2 ? STEREO : MONO);
            lame_set_num_channels(gf, nv->audio_channels);
            lame_set_in_samplerate(gf, nv->audio_sample_rate);
            lame_set_out_samplerate(gf, nv->audio_sample_rate);

            lame_decode_init();
            lame_init_params(gf);
            lame_print_config(gf);

            lame_set_bWriteVbrTag(gfOut, 0);
            lame_set_quality(gfOut, nv->audio_quality);
            lame_set_brate(gfOut, nv->audio_bitrate_closest);
            lame_set_mode(gfOut, nv->audio_channels == 2 ? STEREO : MONO);
            lame_set_num_channels(gfOut, nv->audio_channels);
            lame_set_in_samplerate(gfOut, nv->audio_sample_rate);
            lame_set_out_samplerate(gfOut, nv->audio_sample_rate);

            lame_init_params(gfOut);
    
            pcmlen = sizeof(short int) * nv->audio_sample_rate * 10;
            pcmlbuffer = (short int *)malloc(pcmlen);
            pcmrbuffer = (short int *)malloc(pcmlen);
            pcmtemp    = (short int *)malloc(pcmlen);

            printf( "PCM buffers length %d bytes, %d shorts\n\n", pcmlen, 
                    pcmlen / sizeof(short int) );

            pcmlen /= sizeof(short int);

            if( !pcmlbuffer || !pcmrbuffer || !pcmtemp ) {
                fprintf( stderr, "Can't allocate PCM buffers!\n" );
                return( NULL );
            }

            QueueEnqueueItem( WritingQueue, (QueueItem_t)frame );
            break;
        case 'E':
            done = TRUE;
            break;
        case 'A':
            if( synched ) {
                item = (linkedFrame *)malloc(sizeof(linkedFrame));
                if( item ) {
                    item->frame = frame;
                    LinkedListAdd( audioList, (LinkedListItem_t *)item,
                                   UNLOCKED, AT_TAIL );
                }
            }
            break;
        case 'V':
            if( synched ) {
                item = (linkedFrame *)malloc(sizeof(linkedFrame));
                if( item ) {
                    item->frame = frame;
                    LinkedListAdd( videoList, (LinkedListItem_t *)item,
                                   UNLOCKED, AT_TAIL );
                }
            }
            break;
        case 'S':
            if( frame->header.comptype == 'V' ) {
                syncFrame = 1;
                synched = 1;
            } else {
                nuv_free_frame( frame );
            }
            break;
        default:
            nuv_free_frame( frame );
            break;
        }


        if( syncFrame || done ) {
            VideoLastPrev = VideoPrev - VideoOffset;
            audRaw = FALSE;

            currAud = audioList->head;
            for(currVid = videoList->head; currVid; currVid = currVid->next) {
                timecode = ((linkedFrame *)currVid)->frame->header.timecode;
                if( (timecode - VideoPrev) > ((VideoFrameTime * 2) + 2) ) {
                    /*
                     * We need to update the offset.  Let's do all the Audio
                     * frames up to the current timestamp with the old offset
                     * first
                     */
                    while( (item = (linkedFrame *)currAud) &&
                           item->frame->header.timecode < timecode) {
                        item->frame->header.timecode -= VideoOffset;
                        item->frame->offset = VideoOffset;
                        if( item->frame->header.comptype != '3' ) {
                            audRaw = TRUE;
                        }
                        currAud = currAud->next;
                    }
                    VideoOffset += (timecode - VideoPrev) + VideoFrameTime;
                }
                VideoPrev = timecode;

                item = (linkedFrame *)currVid;
                item->frame->header.timecode -= VideoOffset;
                item->frame->offset = VideoOffset;
            }

            /* Finish offsetting the audio */
            for(; currAud; currAud = currAud->next) {
                item = (linkedFrame *)currAud;
                item->frame->header.timecode -= VideoOffset;
                item->frame->offset = VideoOffset;
                if( item->frame->header.comptype != '3' ) {
                    audRaw = TRUE;
                }
            }

            lameret = 0;
            bufOffset = 0;

            /* Now output audio */    
            for(currAud = audioList->head; currAud; currAud = audioList->head) {
                item = (linkedFrame *)currAud;

                lameret = lame_decode(item->frame->data, 
                                      item->frame->length,
                                      &pcmlbuffer[bufOffset], 
                                      &pcmrbuffer[bufOffset]);

                if (lameret > 0) {
                    bufOffset += lameret;
                } else if (lameret < 0) {
                    fprintf( stderr, "\nlame decode error %d, exiting\n",
                             lameret );
                    return( NULL );
                }

                LinkedListRemove( audioList, currAud, UNLOCKED );
                if( !audRaw ) {
                    /* MP3 audio */
                    QueueEnqueueItem( WritingQueue, (QueueItem_t)item->frame );
                } else {
                    nuv_free_frame( item->frame );
                }
                free( item );
            }


            if( bufOffset ) {
                AudTime = (float)bufOffset / (float)nv->audio_sample_rate;
                VidTime = (float)(VideoPrev + VideoFrameTime - VideoLastPrev -
                                  VideoOffset) / 1000.0;
                Ratio = AudTime / VidTime;

                if( !audRaw ) {
                    /* MP3 Audio - resynchronize video to the audio */
                    if( Ratio > 1.0 ) {
                        /* 
                         * Push out the next Video frame to adjust for the
                         * audio being longer
                         */
                        VideoOffset -= (int)((AudTime - VidTime) * 1000.0 /
                                             0.97);
                    }
                } else {
                    count = bufOffset;

                    /* Create an audio frame */
                    audFrame = (nuvFrame *)malloc(sizeof(nuvFrame));
                    if( !audFrame ) {
                        fprintf( stderr, "Couldn't allocate a frame, "
                                         "aborting!\n" );
                        return( NULL );
                    }
                    audFrame->header.frametype = 'A';
                    audFrame->header.timecode  = (int)(VideoLastPrev * 1000.0);


                    mp3size = ((count * 5) / 4) + 14400;
                    audFrame->data = (char *)malloc(mp3size);
                    if( !audFrame->data ) {
                        fprintf( stderr, "Couldn't allocate a frame, "
                                         "aborting!\n" );
                        return( NULL );
                    }

                    audFrame->length = 
                        lame_encode_buffer(gfOut, pcmlbuffer, pcmrbuffer, 
                                           count, audFrame->data, mp3size);

                    if( audFrame->length < 0 ) {
                        audFrame->length = 0;
                    }

                    audFrame->length += lame_encode_flush_nogap(gfOut, 
                                        &(audFrame->data[audFrame->length]),
                                        7200);
                    if( audFrame->length < 0 ) {
                        audFrame->length = 0;
                    }
                    QueueEnqueueItem( WritingQueue, (QueueItem_t)audFrame );
                }
            }

            /* Output the video */
            for(currVid = videoList->head; currVid; currVid = videoList->head) {
                item = (linkedFrame *)currVid;
                QueueEnqueueItem( WritingQueue, item->frame );
                LinkedListRemove( videoList, currVid, UNLOCKED );
                free( item );
            }

            /* The S or E frame */
            QueueEnqueueItem( WritingQueue, (QueueItem_t)frame );
            syncFrame  = 0;
        }
    }

    lame_close(gf);

    LinkedListDestroy( audioList );
    LinkedListDestroy( videoList );

    return( NULL );
}


/*
 * vim:ts=4:sw=4:ai:et:si:sts=4
 */
